@@ -34,6 +34,8 @@ use strict;
 
 package FAI;
 
+my %partition_table_deps;
+
 ################################################################################
 #
 # @brief Build the mkfs commands for the partition pointed to by $partition
@@ -42,6 +44,9 @@ package FAI;
 # @param $partition Reference to partition in the config hash
 #
 ################################################################################
+
+my @preserved_raid = ();
+
 sub build_mkfs_commands {
 
   my ($device, $partition) = @_;
@@ -183,11 +188,11 @@ sub set_partition_flag_on_phys_dev {
   my $pre = "exist_$d";
   $pre .= ",cleared2_$disk" if (defined($FAI::configs{"PHY_$disk"}));
   &FAI::push_command( "parted -s $disk set $part_no $t on", $pre, "flag_${t}_$d" );
-  if (defined($FAI::partition_table_deps{$disk}) &&
-    $FAI::partition_table_deps{$disk} ne "") {
-    $FAI::partition_table_deps{$disk} .= ",flag_${t}_$d";
+  if (defined($partition_table_deps{$disk}) &&
+    $partition_table_deps{$disk} ne "") {
+    $partition_table_deps{$disk} .= ",flag_${t}_$d";
   } else {
-    $FAI::partition_table_deps{$disk} = "flag_${t}_$d";
+    $partition_table_deps{$disk} = "flag_${t}_$d";
   }
   return 1;
 }
@@ -280,6 +285,64 @@ sub build_cryptsetup_commands {
 
 }
 
+################################################################################
+#
+# @brief Using the configurations from %FAI::configs, a list of commands is
+# built to create all BTRFS volumes/raids
+#
+################################################################################
+sub build_btrfs_commands {
+  foreach my $config (keys %FAI::configs) { # loop through all configs
+    next unless ($config eq "BTRFS");
+
+    #create BTRFS RAIDs
+    foreach my $id (&numsort(keys %{ $FAI::configs{$config}{volumes} })) {
+    #reference to current btrfs volume
+    my $vol = (\%FAI::configs)->{$config}->{volumes}->{$id};
+
+    #the list of BTRFS-RAID devices
+    my @devs = keys %{ $vol->{devices} };
+    my $raidlevel = $vol->{raidlevel};
+    my $mountpoint = $vol->{mountpoint};
+    my $mountoptions = $vol->{mount_options};
+    ($mountoptions =~ m/subvol=([^,\s]+)/ and my $initial_subvolume= $1) or die "You must define an initial subvolume for your BTRFS RAID";
+    my $btrfscreateopts =  $vol->{btrfscreateopts};
+    defined($btrfscreateopts) or $btrfscreateopts = "";
+    my $createopts = $vol->{createopts};
+    defined($createopts) or $createopts = "";
+    my $pre_req = "pt_complete_/dev/vdd";
+
+    # creates the BTRFS volume/RAID
+    if ($raidlevel eq 'single') {
+          &FAI::push_command("mkfs.btrfs -d single $createopts ". join(" ",@devs),
+                             "$pre_req",
+                             "btrfs_built_raid_$id");
+
+        } else {
+          &FAI::push_command("mkfs.btrfs -d raid$raidlevel $createopts ". join(" ",@devs),
+                             "$pre_req",
+                             "btrfs_built_raid_$id");
+        }
+
+    # initial mount, required to create the initial subvolume
+    &FAI::push_command("mount $devs[0] /mnt",
+                       "btrfs_built_raid_$id",
+                       "btrfs_mounted_$id");
+
+    # creating initial subvolume
+    &FAI::push_command("btrfs subvolume create $btrfscreateopts  /mnt/$initial_subvolume",
+                       "btrfs_mounted_$id",
+                       "btrfs_created_$initial_subvolume");
+
+    # unmounting the device itself
+    &FAI::push_command("umount $devs[0]",
+                       "btrfs_created_$initial_subvolume",
+                       "");
+    }
+  }
+}
+
+
 ################################################################################
 #
 # @brief Using the configurations from %FAI::configs, a list of commands is
@@ -307,7 +370,7 @@ sub build_raid_commands {
 
   foreach my $config (keys %FAI::configs) { # loop through all configs
     # no encrypted, tmpfs, LVM or physical devices here
-    next if ($config eq "CRYPT" || $config eq "TMPFS" || $config =~ /^VG_./ || $config =~ /^PHY_./);
+    next if ($config eq "BTRFS" || $config eq "CRYPT" || $config eq "TMPFS" || $config =~ /^VG_./ || $config =~ /^PHY_./);
     ($config eq "RAID") or &FAI::internal_error("Invalid config $config");
 
     # create all raid devices
@@ -363,6 +426,7 @@ sub build_raid_commands {
         &FAI::push_command(
 	    "mdadm --assemble /dev/md$id " . join(" ", grep(!/^missing$/, @eff_devs)),
 	    "$pre_req", "exist_/dev/md$id");
+        push(@preserved_raid, "/dev/md$id");
 
         # create the filesystem on the volume, if requested
         &FAI::build_mkfs_commands("/dev/md$id",
@@ -660,15 +724,16 @@ sub cleanup_vg {
             next;
           }
         }
+
+        &FAI::push_command( "vgchange -a y $vg",
+          "",
+          "pre_wipe_$vg");
         &FAI::push_command( "wipefs -a /dev/$vg/$lv",
-          "$pre_deps_cl",
+          "pre_wipe_$vg,$pre_deps_cl",
           "wipefs_$vg/$lv");
         &FAI::push_command( "lvremove -f /dev/$vg/$lv",
           "wipefs_$vg/$lv",
           "lv_rm_$vg/$lv,self_cleared_/dev/$vg/$lv");
-        &FAI::push_command( "vgchange -a n $vg",
-          "lv_rm_$vg/$lv,self_cleared_/dev/$vg/$lv",
-          "$vg_setup_pre");
         $vg_setup_pre .= ",lv_rm_$vg/$lv";
       }
     } else {
@@ -677,7 +742,9 @@ sub cleanup_vg {
         (keys %{ $FAI::current_lvm_config{$vg}{volumes} });
     }
     &FAI::push_command("true", $vg_setup_pre, "vg_exists_$vg");
-
+    &FAI::push_command( "vgchange -a n $vg",
+      "",
+      "$vg_setup_pre");
     return 0;
   }
 
@@ -688,8 +755,12 @@ sub cleanup_vg {
       join(",self_cleared_", @{ $FAI::current_dev_children{"/dev/$vg/$lv"} })
         if (defined($FAI::current_dev_children{"/dev/$vg/$lv"}) &&
           scalar(@{ $FAI::current_dev_children{"/dev/$vg/$lv"} }));
+
+    &FAI::push_command( "vgchange -a y $vg",
+      "",
+      "pre_wipe_$vg");
     &FAI::push_command( "wipefs -a /dev/$vg/$lv",
-      "$pre_deps_cl",
+      "pre_wipe_$vg,$pre_deps_cl",
       "wipefs_$vg/$lv");
     &FAI::push_command( "lvremove -f /dev/$vg/$lv",
       "wipefs_$vg/$lv",
@@ -729,13 +800,23 @@ sub build_lvm_commands {
     my $vg = $1;
     my $vg_pre = "vgchange_a_n_VG_$vg";
     my $pre_deps_vgc = "";
+    my $preserved = 0;
 
     $pre_deps_vgc = ",self_cleared_" .
      join(",self_cleared_", @{ $FAI::current_dev_children{$d} })
      if (defined($FAI::current_dev_children{$d}) &&
        scalar(@{ $FAI::current_dev_children{$d} }));
     $pre_deps_vgc =~ s/^,//;
-    &FAI::push_command("vgchange -a n $vg", "$pre_deps_vgc", $vg_pre);
+
+    foreach my $raid (@preserved_raid) {
+      my $tmp_vg = `pvdisplay $raid | grep "VG Name"`;
+      chomp $tmp_vg;
+      $tmp_vg = $1 if $tmp_vg =~ /(\S+)$/;
+      $preserved = 1 if ($tmp_vg eq $vg);
+    }
+    &FAI::push_command("vgchange -a n $vg", "$pre_deps_vgc", $vg_pre)
+      unless $preserved;
+
     $vg_pre .= ",pv_sigs_removed_$vg" if (&FAI::cleanup_vg($vg));
     my $pre_deps_cl = "";
     $pre_deps_cl = ",self_cleared_" .
@@ -748,7 +829,7 @@ sub build_lvm_commands {
   foreach my $config (keys %FAI::configs) {
 
     # no physical devices, RAID, encrypted or tmpfs here
-    next if ($config =~ /^PHY_./ || $config eq "RAID" || $config eq "CRYPT" || $config eq "TMPFS");
+    next if ($config eq "BTRFS" || $config =~ /^PHY_./ || $config eq "RAID" || $config eq "CRYPT" || $config eq "TMPFS");
     ($config =~ /^VG_(.+)$/) or &FAI::internal_error("Invalid config $config");
     next if ($1 eq "--ANY--");
     my $vg = $1; # the volume group
@@ -1196,7 +1277,7 @@ sub setup_partitions {
     if($FAI::configs{$config}{disklabel} eq "gpt-bios" and $boot_disk);
 
   ($prev_id > -1) or &FAI::internal_error("No partitions created");
-  $FAI::partition_table_deps{$disk} = "cleared2_$disk,exist_"
+  $partition_table_deps{$disk} = "cleared2_$disk,exist_"
     . &FAI::make_device_name($disk, $prev_id);
 }
 
@@ -1212,7 +1293,7 @@ sub build_disk_commands {
   # loop through all configs
   foreach my $config ( keys %FAI::configs ) {
     # no RAID, encrypted, tmpfs or LVM devices here
-    next if ($config eq "RAID" || $config eq "CRYPT" || $config eq "TMPFS" || $config =~ /^VG_./);
+    next if ($config eq "BTRFS" || $config eq "RAID" || $config eq "CRYPT" || $config eq "TMPFS" || $config =~ /^VG_./);
     ($config =~ /^PHY_(.+)$/) or &FAI::internal_error("Invalid config $config");
     my $disk = $1; # the device to be configured
 
@@ -1222,21 +1303,21 @@ sub build_disk_commands {
         &FAI::push_command( "true", "",
           "exist_" . &FAI::make_device_name($disk, $part_id) );
         # no partition table operations
-        $FAI::partition_table_deps{$disk} = "";
+        $partition_table_deps{$disk} = "";
       }
     } elsif (defined($FAI::configs{$config}{partitions}{0})) {
       # no partition table operations
-      $FAI::partition_table_deps{$disk} = "";
+      $partition_table_deps{$disk} = "";
    } elsif (defined($FAI::configs{$config}{opts_all}{preserve})) {
      foreach my $part_id (&numsort(keys %{ $FAI::configs{$config}{partitions} })) {
        # all partitions exist
        &FAI::push_command( "true", "",
          "exist_" . &FAI::make_device_name($disk, $part_id) );
        # no partition table operations
-       $FAI::partition_table_deps{$disk} = "";
+       $partition_table_deps{$disk} = "";
      }
      # no changes on this disk
-     $FAI::partition_table_deps{$disk} = "";
+     $partition_table_deps{$disk} = "";
     } else {
       # create partitions on non-virtual configs
       &FAI::setup_partitions($config);
@@ -1316,8 +1397,8 @@ sub restore_partition_table {
 ################################################################################
 sub order_commands {
   # first add partition-table-is-complete
-  &FAI::push_command("true", $FAI::partition_table_deps{$_}, "pt_complete_$_")
-    foreach (keys %FAI::partition_table_deps);
+  &FAI::push_command("true", $partition_table_deps{$_}, "pt_complete_$_")
+    foreach (keys %partition_table_deps);
 
   my @pre_deps = ();
   my $i = 1;
